@@ -6,8 +6,8 @@ from jaaql.exceptions.http_status_exception import HttpStatusException, HTTPStat
 from typing import Optional
 from jaaql.mvc.response import JAAQLResponse
 from collections import Counter
+from os.path import dirname
 from jaaql.utilities import crypt_utils
-import os, errno
 import uuid
 import pyotp
 import qrcode
@@ -42,7 +42,7 @@ QUERY__application_del = "DELETE FROM jaaql__application WHERE name = :name"
 QUERY__application_sel = "SELECT * FROM jaaql__application"
 QUERY__application_count = "SELECT COUNT(*) FROM jaaql__application"
 QUERY__application_upd = "UPDATE jaaql__application SET name = coalesce(:new_name, name), description = coalesce(:new_description, description), url = coalesce(:new_url, url) WHERE name = :name"
-QUERY__database_ins = "INSERT INTO jaaql__database (name, description, port, address, jaaql_name, interface_class) VALUES (:name, :description, :port, :address, :jaaql_name, 'DBPGInterface') RETURNING id"
+QUERY__database_ins = "INSERT INTO jaaql__database (name, description, port, address, jaaql_name, interface_class, is_console_level) VALUES (:name, :description, :port, :address, :jaaql_name, 'DBPGInterface', :is_console_level) RETURNING id"
 QUERY__database_sel = "SELECT * FROM jaaql__database"
 QUERY__database_del = "UPDATE jaaql__database SET deleted = current_timestamp WHERE id = :id AND deleted is null"
 QUERY__database_count = "SELECT COUNT(*) FROM jaaql__database"
@@ -81,7 +81,7 @@ QUERY__log_ins = "INSERT INTO jaaql__log (the_user, occurred, duration_ms, encry
 QUERY__user_log_sel = "SELECT occurred, encrypted_address as address, encrypted_ua as user_agent, status, endpoint, duration_ms, encrypted_exception as exception FROM jaaql__my_logs"
 QUERY__user_log_count = "SELECT COUNT(*) FROM jaaql__my_logs"
 QUERY__my_configs = "SELECT * FROM jaaql__my_configurations"
-QUERY__authorized_configuration = """SELECT parameter_name, parameter_description, username, password, name, port, address, database FROM jaaql__their_authorized_configurations WHERE application = :application AND configuration = :configuration AND pg_has_role(:username, application_role, 'MEMBER') AND pg_has_role(:username, database_role, 'MEMBER')
+QUERY__authorized_configuration = """SELECT parameter_name, parameter_description, username, password, name, port, address, database, is_console_level FROM jaaql__their_authorized_configurations WHERE application = :application AND configuration = :configuration AND pg_has_role(:username, application_role, 'MEMBER') AND pg_has_role(:username, database_role, 'MEMBER')
         AND precedence IN (
             SELECT
                 MAX(jad.precedence)
@@ -114,6 +114,7 @@ KEY__ip = "ip"
 KEY__ua = "ua"
 KEY__status = "status"
 KEY__endpoint = "endpoint"
+KEY__is_console_level = "is_console_level"
 
 DELETION_PURPOSE__account = "account"
 DELETION_PURPOSE__application = "application"
@@ -138,6 +139,8 @@ JWT__created = "created"
 JWT__ua = "ua"
 JWT__ip = "ip"
 
+KEY__db_name = "db_name"
+
 
 class JAAQLModel(BaseJAAQLModel):
 
@@ -150,7 +153,7 @@ class JAAQLModel(BaseJAAQLModel):
         username = username.lower()
 
         inputs = {KEY__username: username}
-        users = self._execute_supplied_statement(self.jaaql_lookup_connection, QUERY__fetch_user_latest_password,
+        users = self.execute_supplied_statement(self.jaaql_lookup_connection, QUERY__fetch_user_latest_password,
                                                  inputs, as_objects=True,
                                                  decrypt_columns=[ATTR__password_hash, KEY__totp_iv],
                                                  encryption_key=self.get_db_crypt_key())
@@ -167,7 +170,7 @@ class JAAQLModel(BaseJAAQLModel):
             ATTR__ip_address: ip_address
         }
 
-        resp = self._execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__user_ip_ins, inputs,
+        resp = self.execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__user_ip_ins, inputs,
                                                           as_objects=True,
                                                           encrypt_parameters=[ATTR__address_hash, ATTR__ip_address],
                                                           encryption_key=self.get_db_crypt_key(),
@@ -186,7 +189,7 @@ class JAAQLModel(BaseJAAQLModel):
                 ATTR__ua: user_agent
             }
 
-            resp = self._execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__user_ua_ins, inputs,
+            resp = self.execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__user_ua_ins, inputs,
                                                               as_objects=True,
                                                               encrypt_parameters=[ATTR__ua_hash, ATTR__ua],
                                                               encryption_key=self.get_db_crypt_key(),
@@ -225,14 +228,23 @@ class JAAQLModel(BaseJAAQLModel):
         if not verified and self.use_mfa:
             raise HttpStatusException(ERR__incorrect_credentials, HTTPStatus.UNAUTHORIZED)
 
-        if last_totp == mfa_key:
+        if last_totp == mfa_key and self.use_mfa:
             raise HttpStatusException(ERR__mfa_reused, HTTPStatus.UNAUTHORIZED)
 
-        self._execute_supplied_statement(self.jaaql_lookup_connection, QUERY__user_totp_upd, {
+        self.execute_supplied_statement(self.jaaql_lookup_connection, QUERY__user_totp_upd, {
             KEY__user_id: user_id,
             KEY__last_totp: mfa_key
         })
 
+    def redeploy(self):
+        f = open(join(dirname(get_jaaql_root()), "redeploy"), "w")
+        f.write("Will be detected and redeployment will now happen")
+        f.close()
+
+        print("Redeploying JAAQL")
+
+        threading.Thread(target=self.exit_jaaql, daemon=True).start()
+
     def authenticate(self, username: str, password: str, mfa_key: str, ip_address: str, user_agent: str,
                      response: JAAQLResponse):
         user, _, _, ip_id, ua_id, last_totp = self.verify_user(username, ip_address, user_agent)
@@ -285,7 +297,8 @@ class JAAQLModel(BaseJAAQLModel):
                 KEY__description: DESCRIPTION__jaaql_db,
                 KEY__port: port,
                 KEY__address: address,
-                KEY__jaaql_name: DATABASE__jaaql_internal_name
+                KEY__jaaql_name: DATABASE__jaaql_internal_name,
+                KEY__is_console_level: False
             }, self.jaaql_lookup_connection)
             self.vault.insert_obj(VAULT_KEY__jaaql_db_id, database_id)
             self.vault.insert_obj(VAULT_KEY__jaaql_lookup_connection, db_connection_string)
@@ -305,16 +318,6 @@ class JAAQLModel(BaseJAAQLModel):
         else:
             raise HttpStatusException(ERR__already_installed)
 
-    def exit_jaaql(self):
-        """
-        Will terminate the worker forcefully which fires a hook in gunicorn_config.py
-        This hook reloads all workers if the file exists in vault
-        :return:
-        """
-        time.sleep(1)
-        open(join(DIR__vault, FILE__was_installed), 'a').close()
-        os._exit(0)
-
     def log(self, user_id: str, occurred: datetime, duration_ms: int, exception: str, contr_input: str, ip: str,
             ua: Optional[str], status: int, endpoint: str, databases: list = None):
         if databases is None:
@@ -332,7 +335,7 @@ class JAAQLModel(BaseJAAQLModel):
             KEY__endpoint: endpoint
         }
 
-        self._execute_supplied_statement(self.jaaql_lookup_connection, QUERY__log_ins, parameters,
+        self.execute_supplied_statement(self.jaaql_lookup_connection, QUERY__log_ins, parameters,
                                          encrypt_parameters=[KEY__exception, KEY__input],
                                          encryption_key=self.get_db_crypt_key())
 
@@ -367,7 +370,7 @@ class JAAQLModel(BaseJAAQLModel):
             KEY__role: username,
             KEY__database: self.vault.get_obj(VAULT_KEY__jaaql_db_id)
         }
-        auth = self._execute_supplied_statement_singleton(self.jaaql_lookup_connection,
+        auth = self.execute_supplied_statement_singleton(self.jaaql_lookup_connection,
                                                           QUERY__database_authorization_sel_one, params,
                                                           as_objects=True,
                                                           decrypt_columns=[KEY__username, KEY__password, KEY__port,
@@ -380,14 +383,14 @@ class JAAQLModel(BaseJAAQLModel):
                                             ), user[KEY__id], ip_id, ua_id, iv, user[ATTR__password_hash], last_totp
 
     def add_application(self, inputs: dict, jaaql_connection: DBInterface):
-        self._execute_supplied_statement(jaaql_connection, QUERY__application_ins, inputs)
+        self.execute_supplied_statement(jaaql_connection, QUERY__application_ins, inputs)
 
     def delete_application(self, inputs: dict):
         return {KEY__deletion_key: self.request_deletion_key(DELETION_PURPOSE__application, inputs)}
 
     def delete_application_confirm(self, inputs: dict, jaaql_connection: DBInterface):
         parameters = self.validate_deletion_key(inputs[KEY__deletion_key], DELETION_PURPOSE__application)
-        self._execute_supplied_statement(jaaql_connection, QUERY__application_del, parameters,
+        self.execute_supplied_statement(jaaql_connection, QUERY__application_del, parameters,
                                          as_objects=True)
 
     def delete_database(self, inputs: dict):
@@ -395,7 +398,7 @@ class JAAQLModel(BaseJAAQLModel):
 
     def delete_database_confirm(self, inputs: dict, jaaql_connection: DBInterface):
         parameters = self.validate_deletion_key(inputs[KEY__deletion_key], DELETION_PURPOSE__database)
-        self._execute_supplied_statement(jaaql_connection, QUERY__database_del, parameters,
+        self.execute_supplied_statement(jaaql_connection, QUERY__database_del, parameters,
                                          as_objects=True)
 
     def get_applications(self, inputs: dict, jaaql_connection: DBInterface):
@@ -406,7 +409,7 @@ class JAAQLModel(BaseJAAQLModel):
                                          where_query, where_parameters)
 
     def add_database(self, inputs: dict, jaaql_connection: DBInterface):
-        return self._execute_supplied_statement(jaaql_connection,
+        return self.execute_supplied_statement(jaaql_connection,
                                                 QUERY__database_ins, inputs, as_objects=True,
                                                 encryption_key=self.get_db_crypt_key(),
                                                 encrypt_parameters=[
@@ -416,7 +419,7 @@ class JAAQLModel(BaseJAAQLModel):
                                                 ])[0][KEY__id]
 
     def update_application(self, inputs: dict, jaaql_connection: DBInterface):
-        self._execute_supplied_statement(jaaql_connection, QUERY__application_upd, inputs)
+        self.execute_supplied_statement(jaaql_connection, QUERY__application_upd, inputs)
 
     def get_databases(self, inputs: dict, jaaql_connection: DBInterface):
         paging_dict, parameters = self.setup_paging_parameters(inputs)
@@ -429,7 +432,7 @@ class JAAQLModel(BaseJAAQLModel):
                                          decrypt_columns=[KEY__port, KEY__address, KEY__database_name])
 
     def add_application_parameter(self, inputs: dict, jaaql_connection: DBInterface):
-        self._execute_supplied_statement(jaaql_connection, QUERY__application_parameter_ins, inputs)
+        self.execute_supplied_statement(jaaql_connection, QUERY__application_parameter_ins, inputs)
 
     def get_application_parameters(self, inputs: dict, jaaql_connection: DBInterface):
         paging_query, where_query, where_parameters, parameters = self.construct_paging_queries(inputs)
@@ -443,11 +446,11 @@ class JAAQLModel(BaseJAAQLModel):
 
     def delete_application_parameter_confirm(self, inputs: dict, jaaql_connection: DBInterface):
         parameters = self.validate_deletion_key(inputs[KEY__deletion_key], DELETION_PURPOSE__application_parameter)
-        self._execute_supplied_statement(jaaql_connection, QUERY__application_parameter_del, parameters,
+        self.execute_supplied_statement(jaaql_connection, QUERY__application_parameter_del, parameters,
                                          as_objects=True)
 
     def add_application_configuration(self, inputs: dict, jaaql_connection: DBInterface):
-        self._execute_supplied_statement(jaaql_connection, QUERY__application_configuration_ins, inputs)
+        self.execute_supplied_statement(jaaql_connection, QUERY__application_configuration_ins, inputs)
 
     def get_application_configurations(self, inputs: dict, jaaql_connection: DBInterface):
         paging_query, where_query, where_parameters, parameters = self.construct_paging_queries(inputs)
@@ -461,11 +464,11 @@ class JAAQLModel(BaseJAAQLModel):
 
     def delete_application_configuration_confirm(self, inputs: dict, jaaql_connection: DBInterface):
         parameters = self.validate_deletion_key(inputs[KEY__deletion_key], DELETION_PURPOSE__application_configuration)
-        self._execute_supplied_statement(jaaql_connection, QUERY__application_configuration_del, parameters,
+        self.execute_supplied_statement(jaaql_connection, QUERY__application_configuration_del, parameters,
                                          as_objects=True)
 
     def add_application_argument(self, inputs: dict, jaaql_connection: DBInterface):
-        self._execute_supplied_statement(jaaql_connection, QUERY__application_argument_ins, inputs)
+        self.execute_supplied_statement(jaaql_connection, QUERY__application_argument_ins, inputs)
 
     def get_application_arguments(self, inputs: dict, jaaql_connection: DBInterface):
         paging_query, where_query, where_parameters, parameters = self.construct_paging_queries(inputs)
@@ -479,11 +482,11 @@ class JAAQLModel(BaseJAAQLModel):
 
     def delete_application_argument_confirm(self, inputs: dict, jaaql_connection: DBInterface):
         parameters = self.validate_deletion_key(inputs[KEY__deletion_key], DELETION_PURPOSE__application_argument)
-        self._execute_supplied_statement(jaaql_connection, QUERY__application_argument_del, parameters,
+        self.execute_supplied_statement(jaaql_connection, QUERY__application_argument_del, parameters,
                                          as_objects=True)
 
     def add_application_authorization(self, inputs: dict, jaaql_connection: DBInterface):
-        self._execute_supplied_statement(jaaql_connection, QUERY__application_authorization_ins, inputs)
+        self.execute_supplied_statement(jaaql_connection, QUERY__application_authorization_ins, inputs)
 
     def get_application_authorizations(self, inputs: dict, jaaql_connection: DBInterface):
         paging_query, where_query, where_parameters, parameters = self.construct_paging_queries(inputs)
@@ -497,13 +500,13 @@ class JAAQLModel(BaseJAAQLModel):
 
     def delete_application_authorization_confirm(self, inputs: dict, jaaql_connection: DBInterface):
         parameters = self.validate_deletion_key(inputs[KEY__deletion_key], DELETION_PURPOSE__application_authorization)
-        self._execute_supplied_statement(jaaql_connection, QUERY__application_authorization_del, parameters,
-                                         as_objects=True)
+        self.execute_supplied_statement(jaaql_connection, QUERY__application_authorization_del, parameters,
+                                        as_objects=True)
 
     def add_database_authorization(self, inputs: dict, connection: DBInterface):
-        self._execute_supplied_statement(connection, QUERY__database_authorization_ins, inputs,
-                                         encrypt_parameters=[KEY__username, KEY__password],
-                                         encryption_key=self.get_db_crypt_key())
+        self.execute_supplied_statement(connection, QUERY__database_authorization_ins, inputs,
+                                        encrypt_parameters=[KEY__username, KEY__password],
+                                        encryption_key=self.get_db_crypt_key())
 
     def get_database_authorizations(self, inputs: dict, jaaql_connection: DBInterface):
         paging_dict, parameters = self.setup_paging_parameters(inputs)
@@ -519,10 +522,11 @@ class JAAQLModel(BaseJAAQLModel):
 
     def delete_database_authorization_confirm(self, inputs: dict, jaaql_connection: DBInterface):
         parameters = self.validate_deletion_key(inputs[KEY__deletion_key], DELETION_PURPOSE__database_authorization)
-        self._execute_supplied_statement(jaaql_connection, QUERY__database_authorization_del, parameters,
+        self.execute_supplied_statement(jaaql_connection, QUERY__database_authorization_del, parameters,
                                          as_objects=True)
 
     def user_invite(self, inputs: dict):
+        # TODO Make sure that we can't invite a user that already exists
         ms_two_weeks = 1000 * 60 * 60 * 24 * 14
         return crypt_utils.jwt_encode(self.vault.get_obj(VAULT_KEY__jwt_crypt_key), inputs, ms_two_weeks)
 
@@ -556,7 +560,7 @@ class JAAQLModel(BaseJAAQLModel):
             ATTR__mobile: mobile,
             KEY__totp_iv: totp_iv
         }
-        user_id = self._execute_supplied_statement(jaaql_connection, QUERY__user_ins, parameters, as_objects=True,
+        user_id = self.execute_supplied_statement(jaaql_connection, QUERY__user_ins, parameters, as_objects=True,
                                                    encrypt_parameters=[KEY__totp_iv],
                                                    encryption_key=self.get_db_crypt_key())
         user_id = user_id[0][KEY__id]
@@ -566,7 +570,7 @@ class JAAQLModel(BaseJAAQLModel):
             ATTR__the_user: user_id,
             ATTR__password_hash: password
         }
-        self._execute_supplied_statement(jaaql_connection, QUERY__user_password_ins, parameters,
+        self.execute_supplied_statement(jaaql_connection, QUERY__user_password_ins, parameters,
                                          encrypt_parameters=[ATTR__password_hash],
                                          encryption_key=self.get_db_crypt_key(),
                                          encryption_salts={ATTR__password_hash: user_id})
@@ -589,7 +593,7 @@ class JAAQLModel(BaseJAAQLModel):
             KEY__username: username,
             KEY__password: jaaql_password
         }
-        self._execute_supplied_statement(self.jaaql_lookup_connection, QUERY__user_create_role, inputs)
+        self.execute_supplied_statement(self.jaaql_lookup_connection, QUERY__user_create_role, inputs)
 
         qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L, box_size=10, border=4)
         qr.add_data(totp_uri)
@@ -624,7 +628,7 @@ class JAAQLModel(BaseJAAQLModel):
                                          encryption_key=self.get_db_crypt_key())
 
     def my_configs(self, jaaql_connection: DBInterface):
-        return self._execute_supplied_statement(jaaql_connection, QUERY__my_configs, as_objects=True)
+        return self.execute_supplied_statement(jaaql_connection, QUERY__my_configs, as_objects=True)
 
     def change_password(self, http_inputs: dict, totp_iv: str, oauth_token: str, password_hash: str, user_id: str,
                         last_totp: str, jaaql_connection: DBInterface):
@@ -648,7 +652,7 @@ class JAAQLModel(BaseJAAQLModel):
             ATTR__the_user: user_id,
             ATTR__password_hash: new_password
         }
-        self._execute_supplied_statement(jaaql_connection, QUERY__user_password_ins, parameters,
+        self.execute_supplied_statement(jaaql_connection, QUERY__user_password_ins, parameters,
                                          encrypt_parameters=[KEY__totp_iv],
                                          encryption_key=self.get_db_crypt_key(),
                                          encryption_salts={ATTR__password_hash: user_id})
@@ -670,7 +674,7 @@ class JAAQLModel(BaseJAAQLModel):
 
     def close_account_confirm(self, inputs: dict, user_id: str):
         self.validate_deletion_key(inputs[KEY__deletion_key], DELETION_PURPOSE__account)
-        self._execute_supplied_statement(self.jaaql_lookup_connection, QUERY__user_del, {ATTR__the_user: user_id})
+        self.execute_supplied_statement(self.jaaql_lookup_connection, QUERY__user_del, {ATTR__the_user: user_id})
 
     @staticmethod
     def build_db_addr(row: dict):
@@ -681,7 +685,7 @@ class JAAQLModel(BaseJAAQLModel):
         http_inputs[KEY__username] = jaaql_connection.username
 
         # It is not a mistake to use the jaaql_lookup_connection rather than the user's connection here
-        data = self._execute_supplied_statement(self.jaaql_lookup_connection, QUERY__authorized_configuration,
+        data = self.execute_supplied_statement(self.jaaql_lookup_connection, QUERY__authorized_configuration,
                                                 http_inputs, as_objects=True, encryption_key=self.get_db_crypt_key(),
                                                 decrypt_columns=[
                                                     KEY__username,
@@ -703,7 +707,8 @@ class JAAQLModel(BaseJAAQLModel):
         # Combines the existing row, overriding KEY__database with a JWT containing the encrypted database UUID
         ret = [{KEY__parameter_description: row[KEY__parameter_description], KEY__parameter_name: row[
             KEY__parameter_name], KEY__database: crypt_utils.jwt_encode(
-            jwt_key, {KEY__db_url: crypt_utils.encrypt(obj_key, JAAQLModel.build_db_addr(row))})} for row in data]
+            jwt_key, {KEY__db_url: crypt_utils.encrypt(obj_key, JAAQLModel.build_db_addr(
+                row) + "##" + str(row[KEY__is_console_level]))})} for row in data]
 
         return ret
 
@@ -718,7 +723,16 @@ class JAAQLModel(BaseJAAQLModel):
         else:
             db_url = crypt_utils.jwt_decode(jwt_key, database)[KEY__db_url]
             db_url = crypt_utils.decrypt(obj_key, db_url)
+            db_parts = db_url.split("##")
+            db_url = db_parts[0]
+            is_console_level = db_parts[1] == "True"
             address, port, database, username, password = DBInterface.fracture_uri(db_url)
+
+            if KEY__db_name in http_inputs and is_console_level:
+                database = http_inputs[KEY__db_name]
+            elif KEY__db_name in http_inputs:
+                raise HttpStatusException("Cannot override DB", HTTPStatus.UNAUTHORIZED)
+
             database = DBInterface.create_interface(self.config, address, port, database, username, password)
 
         if KEY__database in http_inputs:
